/*
    libnoxgl - initialize OpenGL without X11 (mainly for off-screen rendering)
    Copyright (C) 2023 Tibor 'Igor2' Palinkas
    Credits: Concepts based on kmscube

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Contact:
      Project page: http://www.repo.hu/projects/libnoxgl
      Lead developer: http://www.repo.hu/projects/libnoxgl/contact.html
*/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <xf86drm.h>
#include <xf86drmMode.h>
#include <gbm.h>
#include <EGL/egl.h>
#include <EGL/eglext.h>

#include <libdrm/drm_fourcc.h>

#include <GL/gl.h>

#include "libnoxgl.h"

typedef struct {
	int fd;
	uint32_t crtc_id; /* not used in off-screen */
	long hdisplay, vdisplay; /* display size */
} drm_t;

static uint32_t find_crtc_for_encoder(const drmModeRes *resources, const drmModeEncoder *encoder)
{
	int n;
	uint32_t mask = 1;


	/* possible_crtcs is a bitmask: https://dvdhrm.wordpress.com/2012/09/13/linux-drm-mode-setting-api */
	for(n = 0; n < resources->count_crtcs; n++, mask = mask << 1)
		if (encoder->possible_crtcs & mask)
			return resources->crtcs[n]; /* this is the crc ID */

	return -1; /* no match */
}


static uint32_t find_crtc_for_connector(const drm_t *drm, const drmModeRes *resources, const drmModeConnector *connector)
{
	int n;

	for(n = 0; n < connector->count_encoders; n++) {
		uint32_t crtc_id;
		drmModeEncoder *encoder = drmModeGetEncoder(drm->fd, connector->encoders[n]);

		if (encoder == NULL) continue;

		drmModeFreeEncoder(encoder);

		crtc_id = find_crtc_for_encoder(resources, encoder);
		if (crtc_id != -1)
			return crtc_id;
	}

	return -1; /* not found */
}

/* Returns the first connected connector (caller needs to free it) or NULL
   if nothing is connected */
static drmModeConnector *find_connected_connector(drm_t *drm, drmModeRes *resources)
{
	int n;

	for(n = 0; n < resources->count_connectors; n++) {
		drmModeConnector *connector = drmModeGetConnector(drm->fd, resources->connectors[n]);
		if (connector->connection == DRM_MODE_CONNECTED)
			return connector;
		drmModeFreeConnector(connector);
	}

	return NULL;
}

/* Find preferred mode or the highest resolution mode and return it
   (or NULL if no usable mode found). */
static drmModeModeInfo *find_best_mode(drm_t *drm, drmModeConnector *connector)
{
	int n;
	long best_area;
	drmModeModeInfo *best_mode;

	best_area = 0;
	best_mode = NULL;

	for(n = 0; n < connector->count_modes; n++) {
		int area;
		drmModeModeInfo *mode = &connector->modes[n];

		/* https://docs.kernel.org/gpu/drm-kms.html says there's only one
		   preferred, typically the native resolution of a tft panel */
		if (mode->type & DRM_MODE_TYPE_PREFERRED)
			return mode;

		area = mode->hdisplay * mode->vdisplay;
		if (area > best_area) {
			best_mode = mode;
			best_area = area;
		}
	}

	return best_mode;
}

/* Returns the encoder for a connector (caller needs to free it) or NULL
   if nothing is connected */
static drmModeEncoder *find_encoder(drm_t *drm, drmModeRes *resources, drmModeConnector *connector)
{
	int n;

	for(n = 0; n < resources->count_encoders; n++) {
		drmModeEncoder *encoder = drmModeGetEncoder(drm->fd, resources->encoders[n]);
		if (encoder->encoder_id == connector->encoder_id)
			return encoder;
		drmModeFreeEncoder(encoder);
	}
	
	return NULL;
}

static int init_drm(drm_t *drm, const char *device, int off_screen)
{
	drmModeRes *resources = NULL;
	drmModeConnector *connector = NULL;
	drmModeEncoder *encoder = NULL;
	drmModeModeInfo *mode;

	drm->fd = open(device, O_RDWR);

	if (drm->fd < 0) {
		fprintf(stderr, "libnoxgl: could not open drm device at '%s'\n", device);
		return -1;
	}

	resources = drmModeGetResources(drm->fd);
	if (resources == NULL) {
		fprintf(stderr, "libnoxgl: drmModeGetResources: %s\n", strerror(errno));
		return -1;
	}

	mode = NULL;
	connector = find_connected_connector(drm, resources);
	if (connector == NULL) {
		static drmModeModeInfo mode_off_screen;
		if (!off_screen) {
			fprintf(stderr, "libnoxgl: drm: no connected connector\n");
			goto error;
		}
		mode_off_screen.hdisplay = mode_off_screen.vdisplay = 32768;
		mode = &mode_off_screen;
	}

	if (mode == NULL) {
		mode = find_best_mode(drm, connector);
		if (mode == NULL) {
			fprintf(stderr, "libnoxgl: drm: could not find mode\n");
			goto error;
		}
	}

	if (!off_screen) {
		/* figure crtc_id using the connector's encoder or the connector as fallback */
		encoder = find_encoder(drm, resources, connector);
		if (encoder == NULL) {
			drm->crtc_id = find_crtc_for_connector(drm, resources, connector);
			if (drm->crtc_id == -1) {
				fprintf(stderr, "libnoxgl: drm: no crtc found\n");
				goto error;
			}
		}
		else
			drm->crtc_id = -1;
	}


	if (encoder != NULL)   drmModeFreeEncoder(encoder);
	if (connector != NULL) drmModeFreeConnector(connector);
	if (resources != NULL) drmModeFreeResources(resources);
	return 0;

	error:;
	if (encoder != NULL)   drmModeFreeEncoder(encoder);
	if (connector != NULL) drmModeFreeConnector(connector);
	if (resources != NULL) drmModeFreeResources(resources);
	return -1;
}

static void uninit_drm(drm_t *drm)
{
	close(drm->fd);
}


typedef struct {
	struct gbm_device *dev;
	struct gbm_surface *surface;
	uint32_t format;
	long width, height;
} gbm_t;

static int init_gbm(gbm_t *gbm, int drm_fd, long w, long h, uint64_t modifier)
{
	gbm->format = GBM_FORMAT_XRGB8888;
	gbm->surface = NULL;
	gbm->dev = gbm_create_device(drm_fd);

	if (gbm->dev == NULL) {
		fprintf(stderr, "libnoxgl: gbm failed to create device\n");
		return -1;
	}

	if (gbm_surface_create_with_modifiers != NULL)
		gbm->surface = gbm_surface_create_with_modifiers(gbm->dev, w, h, gbm->format, &modifier, 1);

	if (gbm->surface == NULL) {
		if (modifier != DRM_FORMAT_MOD_LINEAR) {
			fprintf(stderr, "libnoxgl: gbm modifiers not supported\n");
			return -1;
		}
		gbm->surface = gbm_surface_create(gbm->dev, w, h, gbm->format, GBM_BO_USE_SCANOUT | GBM_BO_USE_RENDERING);
	}

	if (!gbm->surface) {
		fprintf(stderr, "libnoxgl: failed to create gbm surface\n");
		return -1;
	}

	gbm->width = w;
	gbm->height = h;

	return 0;
}

static void uninit_gbm(gbm_t *gbm, int drm_fd)
{
	gbm_surface_destroy(gbm->surface);
	gbm_device_destroy(gbm->dev);
}


/* Return whether ext is in extension_list (1 or 0, bool) */
static int has_ext(const char *extension_list, const char *ext)
{
	const char *curr, *next;
	int elen = strlen(ext);

	if ((extension_list == NULL) || (*extension_list == '\0'))
		return 0;

	for(curr = extension_list; (curr != NULL); curr = next) {
		/* navigate to the start of the coming word */
		while(*curr == ' ') curr++;
		if (*curr == '\0')
			return 0; /* was the last word */

		/* compare from word start */
		if ((memcmp(curr, ext, elen) == 0) && ((curr[elen] == ' ') || (curr[elen] == '\0')))
			return 1;

		/* skip to the next word; leading space will be trimmed in the next iteration */
		next = strchr(curr, ' ');
	}

	return 0;
}

typedef struct {
	EGLDisplay display;
	EGLConfig config;

	PFNEGLGETPLATFORMDISPLAYEXTPROC eglGetPlatformDisplayEXT;

	int modifiers_supported;

	EGLint major, minor;
	const char *exts_client, *exts_dpy, *exts_gl;
} egl_t;


static int find_config_idx_for_visual(EGLDisplay egl_display, EGLint visual_id, EGLConfig *configs, int num_cfgs)
{
	int n;

	for(n = 0; n < num_cfgs; n++) {
		EGLint id;

		if (!eglGetConfigAttrib(egl_display, configs[n], EGL_NATIVE_VISUAL_ID, &id))
			continue;

		if (id == visual_id)
			return n;
	}

	return -1;
}

/* return 1 on success */
static int egl_choose_config(EGLDisplay egl_display, const EGLint *attribs, EGLint visual_id, EGLConfig *config_out)
{
	EGLint num_cfgs = 0, matched = 0;
	EGLConfig *configs;
	int config_index = -1;

	memset(config_out, 0, sizeof(EGLConfig));

	/* figure the number of configs */
	if (!eglGetConfigs(egl_display, NULL, 0, &num_cfgs) || (num_cfgs < 1)) {
		fprintf(stderr, "libnoxgl: no EGL configs to choose from\n");
		return 0;
	}

	configs = malloc(num_cfgs * sizeof(EGLConfig));
	if (configs == NULL)
		return 0;

	if (!eglChooseConfig(egl_display, attribs, configs, num_cfgs, &matched) || !matched) {
		fprintf(stderr, "libnoxgl: no EGL configs with appropriate attributes\n");
		goto out;
	}

	if (!visual_id)
		config_index = 0;

	if (config_index == -1)
		config_index = find_config_idx_for_visual(egl_display, visual_id, configs, matched);

	if (config_index != -1)
		*config_out = configs[config_index];

out:
	free(configs);
	if (config_index == -1)
		return 0;

	return 1;
}

#define get_proc_client(ext, name) \
	do { \
		if (has_ext(egl->exts_client, #ext)) \
			egl->name = (void *)eglGetProcAddress(#name); \
	} while(0)

#define get_proc_dpy(ext, name) \
	do { \
		if (has_ext(egl->exts_dpy, #ext)) \
			egl->name = (void *)eglGetProcAddress(#name); \
	} while(0)

#define get_proc_gl(ext, name) \
	do { \
		if (has_ext(egl->exts_gl, #ext)) \
			egl->name = (void *)eglGetProcAddress(#name); \
	} while(0)

static int init_egl(egl_t *egl, const gbm_t *gbm, int samples, int off_screen)
{
	const EGLint config_attribs[] = {
		EGL_SURFACE_TYPE, (off_screen ? 0 : EGL_WINDOW_BIT),
		EGL_RED_SIZE, 1,
		EGL_GREEN_SIZE, 1,
		EGL_BLUE_SIZE, 1,
		EGL_ALPHA_SIZE, 0,
		EGL_RENDERABLE_TYPE, EGL_OPENGL_BIT,
		EGL_CONFORMANT, EGL_OPENGL_BIT,
		EGL_SAMPLES, samples,
		EGL_NONE
	};

	egl->exts_client = eglQueryString(EGL_NO_DISPLAY, EGL_EXTENSIONS);
	get_proc_client(EGL_EXT_platform_base, eglGetPlatformDisplayEXT);

	if (egl->eglGetPlatformDisplayEXT)
		egl->display = egl->eglGetPlatformDisplayEXT(EGL_PLATFORM_GBM_KHR, gbm->dev, NULL);
	else
		egl->display = eglGetDisplay((void *)gbm->dev);

	if (!eglInitialize(egl->display, &egl->major, &egl->minor)) {
		fprintf(stderr, "libnoxgl: egl: failed to initialize\n");
		return -1;
	}

	egl->exts_dpy = eglQueryString(egl->display, EGL_EXTENSIONS);

	egl->modifiers_supported = has_ext(egl->exts_dpy, "EGL_EXT_image_dma_buf_import_modifiers");

	if (!eglBindAPI(EGL_OPENGL_API)) {
		fprintf(stderr, "libnoxgl: failed to bind api EGL_OPENGL_API\n");
		return -1;
	}

	if (!egl_choose_config(egl->display, config_attribs, gbm->format, &egl->config)) {
		fprintf(stderr, "libnoxgl: failed to choose config\n");
		return -1;
	}

	return 0;
}

static void uninit_egl(egl_t *egl, const gbm_t *gbm)
{
	eglTerminate(egl->display);
}


static void make_current_egl(egl_t *egl, void *ctx, void *surface)
{
	eglMakeCurrent(egl->display, surface, surface, ctx);
}

static void print_info_egl(egl_t *egl)
{
	printf("Using display %p with EGL version %d.%d\n", egl->display, egl->major, egl->minor);

	printf("===================================\n");
	printf("EGL information:\n");
	printf("  version: \"%s\"\n", eglQueryString(egl->display, EGL_VERSION));
	printf("  vendor: \"%s\"\n", eglQueryString(egl->display, EGL_VENDOR));
	printf("  client extensions: \"%s\"\n", egl->exts_client);
	printf("  display extensions: \"%s\"\n", egl->exts_dpy);
	printf("===================================\n");

	printf("OpenGL 2.x information:\n");
	printf("  version: \"%s\"\n", glGetString(GL_VERSION));
	printf("  shading language version: \"%s\"\n", glGetString(GL_SHADING_LANGUAGE_VERSION));
	printf("  vendor: \"%s\"\n", glGetString(GL_VENDOR));
	printf("  renderer: \"%s\"\n", glGetString(GL_RENDERER));
	printf("  extensions: \"%s\"\n", egl->exts_gl);
	printf("===================================\n");
}

static int create_ctx_egl(void **ctx_out, void **surface_out, egl_t *egl, const gbm_t *gbm)
{
	static const EGLint context_attribs[] = {
		EGL_CONTEXT_MAJOR_VERSION, 3,
/*		EGL_CONTEXT_MINOR_VERSION, 0,  <--- this is too narrow */
/*		EGL_CONTEXT_OPENGL_FORWARD_COMPATIBLE, EGL_TRUE, <--- this breaks*/
		EGL_NONE
	};

	*ctx_out = eglCreateContext(egl->display, egl->config, EGL_NO_CONTEXT, context_attribs);
	if (*ctx_out == NULL) {
		fprintf(stderr, "libnoxgl: failed to create context\n");
		return -1;
	}

	*surface_out = eglCreateWindowSurface(egl->display, egl->config, (EGLNativeWindowType)gbm->surface, NULL);
	if (*surface_out == EGL_NO_SURFACE) {
		fprintf(stderr, "libnoxgl: failed to create egl surface\n");
		return -1;
	}

	make_current_egl(egl, *ctx_out, *surface_out);

	egl->exts_gl = (char *)glGetString(GL_EXTENSIONS);

	return 0;
}

static void destroy_ctx_egl(void *ctx, void *surface, egl_t *egl, const gbm_t *gbm)
{
	eglDestroySurface(egl->display, surface);
	eglDestroyContext(egl->display, ctx);
}


/*** API ***/

const int noxgl_ver_major = VER_MAJOR;
const int noxgl_ver_minor = VER_MINOR;
const int noxgl_ver_patch = VER_PATCH;

struct noxgl_s {
	drm_t drm;
	gbm_t gbm;
	egl_t egl;
};

noxgl_t *noxgl_init(const char *device, int off_screen)
{
	noxgl_t *noxgl;
	uint64_t modifier = DRM_FORMAT_MOD_LINEAR;
	const int samples = 0;
	int res;

	noxgl = calloc(sizeof(noxgl_t), 1);
	if (noxgl == NULL)
		return NULL;

	if (init_drm(&noxgl->drm, device, off_screen) != 0) {
		fprintf(stderr, "libnoxgl: failed to initialize drm\n");
		free(noxgl);
		return NULL;
	}

	res = init_gbm(&noxgl->gbm, noxgl->drm.fd, noxgl->drm.hdisplay, noxgl->drm.vdisplay, modifier);
	if (res != 0) {
		fprintf(stderr, "libnoxgl: failed to initialize GBM\n");
		/* TODO: uninit drm */
		free(noxgl);
		return NULL;
	}

	res = init_egl(&noxgl->egl, &noxgl->gbm, samples, off_screen);
	if (res != 0) {
		fprintf(stderr, "libnoxgl: failed to initialize EGL\n");
		/* TODO: uninit drm */
		/* TODO: uninit gbm */
		free(noxgl);
		return NULL;
	}

	return noxgl;
}

void noxgl_uninit(noxgl_t *noxgl)
{
	uninit_egl(&noxgl->egl, &noxgl->gbm);
	uninit_gbm(&noxgl->gbm, noxgl->drm.fd);
	uninit_drm(&noxgl->drm);
	free(noxgl);
}


int noxgl_create_context(noxgl_t *noxgl, void **ctx_out, void **surface_out)
{
	return create_ctx_egl(ctx_out, surface_out, &noxgl->egl, &noxgl->gbm);
}

void noxgl_destroy_context(noxgl_t *noxgl, void *ctx, void *surface)
{
	destroy_ctx_egl(ctx, surface, &noxgl->egl, &noxgl->gbm);
}

void noxgl_make_current_context(noxgl_t *noxgl, void *ctx, void *surface)
{
	make_current_egl(&noxgl->egl, ctx, surface);
}

void noxgl_print_info(noxgl_t *noxgl)
{
	print_info_egl(&noxgl->egl);
}

